#include "../../std.h"
#include <iostream>
#include <regex.h>

#include "../../misc/escaped.h"
#include "../../misc/thread.h"
#include "../../misc/log_time.h"
#include "files.h"
#include "se_ns.h"

// HACK
extern string base_gsi_url;
extern string base_ssl_url;


// TODO: use special pointers instead of acquire/release


#undef SOAP_FMAC1
#undef SOAP_FMAC3
#undef SOAP_FMAC5
#undef SOAP_NMAC
#define WITH_NONAMESPACES
#define WITH_NOGLOBAL
#include "stdsoap2.h"
#undef SOAP_FMAC1
#undef SOAP_FMAC3
#undef SOAP_FMAC5
#undef SOAP_NMAC
#define SOAP_FMAC1 static
#define SOAP_FMAC3 static
#define SOAP_FMAC5 static
#define SOAP_NMAC static
#include "file_soap.nsmap"
#include "file_soapC.cpp"
#include "file_soapServer.cpp"

#define __SE_INTERNAL_INCLUDE__
#include "../se.h"

ns__fileinfo::ns__fileinfo() {
  id=NULL; size=NULL; checksum=NULL;
  acl=NULL; created=NULL; state=NULL;
  __size_url=0; url=NULL;
}
ns__fileinfo::~ns__fileinfo() {
}

static bool merge_urls(string &cfg_url,const char* contact_base) {
  if((cfg_url.length() == 0) || (cfg_url[0] == '/')) { // path
    const char* u_ = contact_base;
    u_=strchr(u_,':'); if(u_==NULL) return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; u_=strchr(u_,'/'); if(u_==NULL) u_=contact_base+strlen(contact_base);
    string url(contact_base,u_-contact_base);
    cfg_url=url+cfg_url;
  } else if(cfg_url[0] == ':') { // port+path
    const char* u_ = contact_base;
    u_=strchr(u_,':'); if(u_==NULL) return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; if((*u_) != '/') return false; // bad url
    u_++; u_=strchr(u_,'/'); if(u_==NULL) u_=contact_base+strlen(contact_base);
    int p = cfg_url.find('/'); if(p == string::npos) p=cfg_url.length();
    string url(contact_base,u_-contact_base);
    url+=cfg_url.c_str()+p;
    cfg_url=url;
  } else { // url

  };
  return true;
}

class HTTP_SE_Handle {
 friend class HTTP_SE;
 friend bool se_service_configurator(istream& f,const char* uri,HTTP_Service_Properties &prop);
 friend HTTP_Service* se_service_creator(HTTPS_Connector& c,const char* uri,void* arg);
 private:
  // URL used to announce available files
  string base_url;
  // files handled by this service
  SEFiles files;
  list<string> acl_create;
  list<string> acl_replicate;
  list<string> acl_read;
 public:
  HTTP_SE_Handle(const char* dirpath,const char* url):files(dirpath),base_url(url) { };
  ~HTTP_SE_Handle(void) { };
  operator bool(void) { return (bool)(files); };
  bool register_immediately(void) { return files.register_immediately(); };
  bool register_retry(void) { return files.register_retry(); };
  SENameServer* NS(void) { return files.NS(); };
};

HTTP_Service* se_service_creator(HTTPS_Connector& c,const char* uri,void* arg) {
  HTTP_SE_Handle* handle = (HTTP_SE_Handle*)arg;
  if(!handle) return NULL;
  bool acl_create = false;
  bool acl_replicate = false;
  bool acl_read = false;
  list<AuthEvaluator*>& auths = c.authorizations();
  for(list<AuthEvaluator*>::iterator ii=auths.begin();ii!=auths.end();++ii){
    if(acl_create && acl_replicate && acl_read) break;

    bool for_acl_create=false;
    bool for_acl_replicate=false;
    bool for_acl_read=false;

    if(!acl_create)for(list<string>::iterator i=
           handle->acl_create.begin();i!=handle->acl_create.end();++i) {
      if((*(*ii)) == (*i)) { for_acl_create=true; break; };
    };
    if(!acl_replicate) for(list<string>::iterator i=
           handle->acl_replicate.begin();i!=handle->acl_replicate.end();++i) {
      if((*(*ii)) == (*i)) { for_acl_replicate=true; break; };
    };
    if(!acl_read) for(list<string>::iterator i=
           handle->acl_read.begin();i!=handle->acl_read.end();++i) {
      if((*(*ii)) == (*i)) { for_acl_read=true; break; };
    };
    if(for_acl_read || for_acl_replicate || for_acl_create) {
      if((*(*ii)).evaluate(c.identity()) == AAA_POSITIVE_MATCH) {
        if(for_acl_create) {
          odilog(1,c.pid)<<"User is granted 'create' access through acl '"<<
               (*(*ii)).get_name()<<"'"<<endl; acl_create=true;
        };
        if(for_acl_replicate) {
          odilog(1,c.pid)<<"User is granted 'replicate' access through acl '"<<
               (*(*ii)).get_name()<<"'"<<endl; acl_replicate=true;
        };
        if(for_acl_read) {
          odilog(1,c.pid)<<"User is granted 'read' access through acl '"<<
               (*(*ii)).get_name()<<"'"<<endl; acl_read=true;
        };
      };
    };
  };
  HTTP_SE* h = new HTTP_SE(&c,handle,uri,acl_create,acl_replicate,acl_read);
  return h;
}

class SEFilesList {
 private:
  list<SEFiles*> files;
  LockSimple lock;
 public:
  typedef list<SEFiles*>::iterator iterator;
  void add_files(SEFiles* f);
  SEFilesList(void) { };
  ~SEFilesList(void) { };
  void block(void) { lock.block(); };
  void unblock(void) { lock.unblock(); };
  list<SEFiles*>::iterator begin(void) { return files.begin(); };
  list<SEFiles*>::iterator end(void) { return files.end(); };
};

class Collector_Thread: public Thread {
 private:
  SEFilesList& files;
  virtual void func(void);
 public:
  Collector_Thread(SEFilesList& f):files(f) { };
  ~Collector_Thread(void) { };
};

class Replicator_Thread: public Thread {
 private:
  SEFilesList& files;
  virtual void func(void);
 public:
  Replicator_Thread(void);
  ~Replicator_Thread(void);
};

class Registrator_Thread: public Thread {
 private:
  SEFilesList& files;
  virtual void func(void);
 public:
  Registrator_Thread(void);
  ~Registrator_Thread(void);
};

class HTTP_SE_Thread: public Thread {
 private:
  SEFilesList& files;
  virtual void func(void);
 public:
  HTTP_SE_Thread(SEFilesList& f):files(f) { };
  ~HTTP_SE_Thread(void) { };
};

void SEFilesList::add_files(SEFiles* f) {
  lock.block();
  files.push_back(f);
  lock.unblock();
}

void HTTP_SE_Thread::func(void) {
cerr<<"HTTP_SE_Thread::func"<<endl;     
  while(true) {
    files.block();
    // go through all storages
cerr<<"HTTP_SE_Thread::func: awake"<<endl;
    for(SEFilesList::iterator fs = files.begin();fs!=files.end();++fs) {
      SENameServer* ns = (*fs)->NS();
      files.unblock();
      if(ns) ns->Maintain();
      // go through all files
      // check if it is time to retry registration (TODO)
cerr<<"HTTP_SE_Thread::func: Register"<<endl;
      if(*fs) (*fs)->Register();
      if(*fs) (*fs)->Unregister();
      files.block();
    };
    files.unblock();
cerr<<"HTTP_SE_Thread::func: sleep"<<endl;
    idle(10*60*1000); // 10 minutes
  };
}

// Watch for hanged files uploaded by client
void Collector_Thread::func(void) {
cerr<<"Collector_Thread::func"<<endl;     
  while(true) {
    files.block();
    // go through all storages
cerr<<"Collector_Thread::func: awake"<<endl;
    for(SEFilesList::iterator fs = files.begin();fs!=files.end();++fs) {
      files.unblock();
      if(*fs) (*fs)->RemoveStuck();
      files.block();
    };
    files.unblock();
cerr<<"Collector_Thread::func: sleep"<<endl;
    idle(10*60*1000); // 10 minutes
  };
}

static SEFilesList files_list;
static HTTP_SE_Thread* se_thread = NULL;
static Collector_Thread* collector_thread = NULL;

HTTP_SE::HTTP_SE(HTTPS_Connector *c_,HTTP_SE_Handle *h_,const char* uri_,bool create,bool replicate,bool read):HTTP_ServiceAdv(c_),handle(h_),url(uri_),acl_create(create),acl_replicate(replicate),acl_read(read) {
  odlog(0)<<"SE: new service: service URL: "<<uri_<<endl;
  odlog(0)<<"SE: new service: contacted URL: "<<c_->url()<<endl;
  odlog(0)<<"SE: new service: configured URL: "<<h_->base_url<<endl;
  if(h_->base_url.length()) url=h_->base_url;
  merge_urls(url,c_->url());
  odlog(0)<<"SE: new service: files URL: "<<url<<endl;
  soap_init();
  sp.namespaces=namespaces;
  sp.user=this;
}

HTTP_SE::~HTTP_SE(void) {
  soap_deinit();
}

SEFiles* HTTP_SE::files(void) {
  if(handle==NULL) return NULL;
  return &(handle->files);
}

/*
 * HTTP methods
 */
HTTP_Error HTTP_SE::get(const char* uri,int &keep_alive) {
  if((!c) || (!handle)) {
    return HTTP_NOT_IMPLEMENTED;
  };
  HTTP_Error err;
  if((err=parse_header(keep_alive)) != HTTP_OK) return err;
  if(failure_parsing || unsupported_option_passed) {
    keep_alive=0;
    return HTTP_FAILURE;
  };
  // obtain requested name
  // const char* name = uri+handle->base_url.length();
  const char* name = uri+url.length();
  if(*name != 0) name++; // skip slash
  cerr<<"HTTP_SE::get: name: "<<name<<endl;
/*
  // get file
  int acl_flags = current_file->check_acl(c->identity());
  if(!(acl_flags & FILE_ACC_READ)) {
cerr<<"file not allowed"<<endl;
    if(send_error_response(*c,keep_alive,404,NULL,"404 Not Found") != 0) {
      keep_alive=1;
      return HTTP_ERROR;
    };
    return HTTP_OK;
  };
*/
  SEFileHandle* h = SEFileHandle::open(name,true,handle->files);
  if(h==NULL) {
cerr<<"file not found"<<endl;
    if(send_error_response(*c,keep_alive,404,NULL,"404 Not Found") != 0) {
      keep_alive=1;
      return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  int acl_flags = h->instance().check_acl(c->identity());
  if(!(acl_flags & FILE_ACC_READ)) {
cerr<<"file not allowed"<<endl;
    if(send_error_response(*c,keep_alive,404,NULL,"404 Not Found") != 0) {
      keep_alive=1;
      return HTTP_ERROR;
    };
    return HTTP_OK;
  }; 
  // fix ranges
  uint64_t range_start_,range_end_;
  if(range_passed) {
cerr<<"range_passed"<<endl;
    int n;  
    for(n=0;n<nranges;n++) { // find first suitable range
      if(range_start[n] <= range_end[n]) {
        if(range_start[n] < h->size()) {
          range_end[n]++;
          if(range_end[n] > h->size()) range_end[n]=h->size();
          break;
        };
      } else {
        if(range_start[n] > h->size()) range_start[n]=h->size();
        range_start[n]=h->size()-range_start[n];
        range_end[n]=h->size();
        break;
      };
    };
    if(n<nranges) { range_start_=range_start[n]; range_end_=range_end[n]; }
    else { range_start_=0; range_end_=0; }; // no such data
    h->seek(range_start_);
  } else {
cerr<<"range_not_passed"<<endl;
    range_start_=0; range_end_=h->size();
  };
cerr<<"range_start: "<<range_start_<<endl;
cerr<<"range_end: "<<range_end_<<endl;
cerr<<"size: "<<h->size()<<endl;
  // send reply header
  if(range_end_ <= range_start_) {
cerr<<"range_end <= range_start"<<endl;
    delete h;
    if(send_error_response(*c,keep_alive,416,NULL,"Error 416") != 0) {
cerr<<"failed to send 416"<<endl;
      keep_alive=0;
      return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  bool partial=false;
  if((range_start_ != 0) || (range_end_ != h->size())) partial=true;
cerr<<"partial: "<<partial<<endl;
  err=send_header(keep_alive,range_start_,range_end_,partial,h->size());
cerr<<"send_header: "<<err<<endl;
  if(err != HTTP_OK) { keep_alive=0; delete h; return HTTP_ERROR; };
  // pass requested range to client
  err=HTTP_OK;
  for(;;) {
    if(h->tell() >= range_end_) break; 
    uint64_t l = range_end_-h->tell();
    char buf[65536];
    if(l > sizeof(buf)) l=sizeof(buf);
    l=h->read(buf,l);
    if(l==0) break; // eof or error
    if(c->write(buf,l) != 0) {
      cerr<<"HTTP_SE::get: failure while reading data from disc"<<endl;
      keep_alive=0;
      err=HTTP_FAILURE;
    };
  };
  if(err==HTTP_OK) if(h->tell() < range_end_) { // something wrong with file
    cerr<<"HTTP_SE::get: failure while reading data from disc"<<endl;
    keep_alive=0;
    err=HTTP_FAILURE;
  };
  delete h;
cerr<<"HTTP_SE: normal exit: return"<<endl;
  return err;
}

HTTP_Error HTTP_SE::put(const char* uri,int &keep_alive) {
  if((!c) || (!handle)) {
    return HTTP_NOT_IMPLEMENTED;
  };
  HTTP_Error err;
  if((err=parse_header(keep_alive)) != HTTP_OK) return err;
  if(failure_parsing || unsupported_option_passed) {
    cerr<<"HTTP_SE::put: bad header"<<endl;
    keep_alive=0;
    if(send_header(keep_alive,406) != HTTP_OK) keep_alive=0;
    return HTTP_FAILURE;
  };
  // obtain requested name
  // const char* name = uri+handle->base_url.length();
  const char* name = uri+url.length();
  if(*name != 0) name++; // skip slash
  cerr<<"HTTP_SE::put: name: "<<name<<endl;
  // check state (not yet!)



  // open file
  SEFileHandle* h = SEFileHandle::open(name,false,handle->files);
  if(h==NULL) {
    cerr<<"HTTP_SE::put: file not found: "<<name<<endl;
    if(send_error_response(*c,keep_alive,404,NULL,"404 Not Found") != 0) {
      keep_alive=0;
      return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  int acl_flags = h->instance().check_acl(c->identity());
  if(!(acl_flags & FILE_ACC_WRITE)) {
    cerr<<"file not allowed"<<endl;
    if(send_error_response(*c,keep_alive,404,NULL,"404 Not Found") != 0) {
      keep_alive=1;
      return HTTP_ERROR;
    };
    return HTTP_OK;
  };
  uint64_t range_start_,range_end_;
  if(entity_range_passed) {
    range_start_=entity_range_start;
    range_end_=entity_range_end+1;
    cerr<<"HTTP_SE::put: range_passed: "<<range_start_<<"-"<<range_end_<<endl;
    if(length_passed && ((range_end_-range_start_) != length)) { 
      keep_alive=0; delete h;
      if(send_header(keep_alive,406) != HTTP_OK) keep_alive=0;
      return HTTP_FAILURE;
    };
  } else {
    cerr<<"HTTP_SE::put: range_not_passed"<<endl;
    range_start_=0; range_end_=h->size();
    if(length_passed) range_end_=length;
  };
  if(range_end_ > h->size()) { 
    keep_alive=0; delete h;
    if(send_header(keep_alive,500) != HTTP_OK) keep_alive=0;
    return HTTP_FAILURE;
  };
  h->seek(range_start_);
  err=HTTP_OK;
  // receive entity from network and store locally 
  for(;;) {
    if(h->tell() >= range_end_) break; 
    uint64_t l = range_end_-h->tell();
    char buf[1024*1024]; // 1MB - is it enough ?
    if(l > sizeof(buf)) l=sizeof(buf);
    l=c->read(buf,l);
    if(l == 0) break; // EOF
    l=h->write(buf,l);
    if(l==0) { // error
      odlog(0)<<"Failed to write to file"<<endl;
      err=HTTP_FAILURE;
      break;
    };
  };
  if((err!=HTTP_OK) || (h->tell() < range_end_)) { // something wrong with file
    cerr<<"HTTP_SE::put: failure while reading/writing data from/to network/disc"<<endl;
    keep_alive=0;
    if(send_header(keep_alive,500) != HTTP_OK) keep_alive=0;
    err=HTTP_FAILURE;
  };
  h->close();
  if(h->full()) {                                   // full content stored
    SEFile& f = h->instance();
    f.acquire();
    if((f.complete()) &&                            // metadata complete
       (f.state_reg() == REG_STATE_LOCAL) &&        // not registered yet
       (!(handle->register_immediately()))) {       // register after transfer
          // register new file
      if(f.state_reg(REG_STATE_REGISTERING)) {
        if(handle->NS()->Register(f) != 0) {
          if(handle->register_retry()) {
            cerr<<"Failed to register (will retry)"<<endl;
          } else {
            cerr<<"Failed to register (what todo ?)"<<endl;
          };
          f.state_reg(REG_STATE_LOCAL);
        } else {
          f.state_reg(REG_STATE_ANNOUNCED);
        };
      } else {
        if(handle->register_retry()) {
          cerr<<"Failed to set REGISTERING state (will retry)"<<endl;
        } else {
          cerr<<"Failed to set REGISTERING state (what to do ?)"<<endl;
        };
      };
    };
    f.state_file(FILE_STATE_COMPLETE);
    f.release();
  };
  delete h;
  if(err==HTTP_OK) if(send_header(keep_alive,200) != HTTP_OK) keep_alive=0;
  return err;
}

HTTP_Error HTTP_SE::post(const char* uri,int &keep_alive) {
  // obtain requested name
cerr<<"post: uri: "<<uri<<endl;
  // current_name=uri+handle->base_url.length();
  current_name=uri+url.length();
  if(*current_name != 0) current_name++; // skip slash
cerr<<"post: name: "<<current_name<<endl;
  // get file
  current_file=handle->files.get(current_name);
  if(current_file==NULL) {
cerr<<"post: file: not found"<<endl;
/*
    skip_header(*c,keep_alive);
    if(send_error_response(*c,keep_alive,404,NULL,"404 Not Found") != 0) {
      keep_alive=0;
      return HTTP_ERROR;
    };
    return HTTP_OK;
*/
  } else {
cerr<<"post: file: "<<current_file->id()<<endl;
  };
  enum HTTP_Error r = soap_post(uri,keep_alive);
  current_file=NULL;
  return r;
}

// called during startup of server (once for every base uri)
bool se_service_configurator(istream& f,const char* uri,HTTP_Service_Properties &prop) {
  string path("");
  string ns_url("");
  string base_url("");
  list<string> acl_create;
  list<string> acl_replicate;
  list<string> acl_read;
  int reg_type = registration_immediate;
  for(;!f.eof();) {
    char buf[1024];
    istream_readline(f,buf,sizeof(buf));
    char* p = buf;
    for(;*p;p++) if(!isspace(*p)) break;
    char* command = p;
    if(*command == '#') continue;
    for(;*p;p++) if(isspace(*p)) break;
    int command_len = p-command;
    if(!command_len) continue;
    if((command_len==10) && (strncmp(command,"acl_create",command_len)==0)) {
      for(;;) {
        string name;
        int n = input_escaped_string(p,name);
        if(n == 0) break;
        p+=n;
        acl_create.push_back(name);
      };
    } else if((command_len==13) && (strncmp(command,"acl_replicate",command_len)==0)) {
      for(;;) {
        string name;
        int n = input_escaped_string(p,name);
        if(n == 0) break;
        p+=n;
        acl_replicate.push_back(name);
      };
    } else if((command_len==8) && (strncmp(command,"acl_read",command_len)==0)) {
      for(;;) {
        string name;
        int n = input_escaped_string(p,name);
        if(n == 0) break;
        p+=n;
        acl_read.push_back(name);
      };
    } else if((command_len==4) && (strncmp(command,"path",command_len)==0)) {
      for(;*p;p++) if(!isspace(*p)) break;
      path=p;
      continue;
    } else if((command_len==7) && (strncmp(command,"storage",command_len)==0)) {
      for(;*p;p++) if(!isspace(*p)) break;
      path=p;
      continue;
    } else if((command_len==2) && (strncmp(command,"ns",command_len)==0)) {
      for(;*p;p++) if(!isspace(*p)) break;
      ns_url=p;
      continue;
    } else if((command_len==3) && (strncmp(command,"url",command_len)==0)) {
      for(;*p;p++) if(!isspace(*p)) break;
      base_url=p;
      continue;
    } else if((command_len==12) && (strncmp(command,"registration",command_len)==0)) {
      // parse options separated by ,
cerr<<"registration options: "<<p<<endl;
      for(;;) {
        for(;*p;p++) if(!isspace(*p)) break;
        if(!(*p)) break;
        char* pc = p;
        for(;*pc;pc++) if((*pc)==',') break;
cerr<<"registration option: "<<p<<endl;
        if(strncmp(p,"immediate",pc-p) == 0) { reg_type|=registration_immediate; }
        else if(strncmp(p,"delayed",pc-p) == 0) { reg_type&=~((unsigned int)(registration_immediate)); }
        else if(strncmp(p,"retry",pc-p) == 0) { reg_type|=registration_retry; }
        else if(strncmp(p,"noretry",pc-p) == 0) { reg_type&=~((unsigned int)(registration_retry)); };
        p=pc; if(*p) p++;
      };
    } else if((command_len==3) && (strncmp(command,"end",command_len)==0)) {
      break;
    } else {
      odlog(0)<<"SE(configure): skipping unknown command: "<<command<<endl;
    };
  };
  prop.arg=NULL;
  if(base_url.length() == 0) base_url=uri;
  odlog(0)<<"SE(configure): filesystem path: "<<path<<endl;
  odlog(0)<<"SE(configure): URL: "<<base_url<<endl;
  HTTP_SE_Handle* handle = NULL;
  if(path.length()) {
    handle = new HTTP_SE_Handle(path.c_str(),base_url.c_str());
    if((*handle)) { prop.arg=handle; }
    else { delete handle; handle=NULL; };
  };
  if(!handle) {
    odlog(0)<<"SE(configure): failed"<<endl;
  } else {
    handle->files.NS(ns_url.c_str(),base_url.c_str(),reg_type);
    handle->acl_create.splice(handle->acl_create.end(),acl_create);
    handle->acl_replicate.splice(handle->acl_create.end(),acl_replicate);
    handle->acl_read.splice(handle->acl_create.end(),acl_read);
    // run thread to handle delayed registration and other things
    // one thread per server
    odlog(0)<<"SE(configure): going to start thread"<<endl;
    if(se_thread == NULL) {
      odlog(0)<<"SE(configure): starting thread"<<endl;
      se_thread = new HTTP_SE_Thread(files_list);
      if(se_thread) se_thread->attention();
    };
    if(collector_thread == NULL) {
      odlog(0)<<"SE(configure): starting thread"<<endl;
      collector_thread = new Collector_Thread(files_list);
      if(collector_thread) collector_thread->attention();
    };
    files_list.add_files(&(handle->files));
  };
  prop.subtree=true;
  odlog(0)<<"SE(configure): serving subtree"<<endl;
  return true;
}

SEFile* HTTP_SE::new_file(SEAttributes& attr) {
  if((!c) || (!handle)) {
    return NULL;
  };
  odlog(0)<<"SE: creating new file: "<<attr.id()<<endl;
  SEFile* f = new SEFile(handle->files.Path(),attr);
  if(f) { if(!(*f)) { delete f; f=NULL; }; };
  if(!f) { odlog(0)<<"SE: file creation failed"<<endl; return NULL; };
  return handle->files.add(*f);
}

#ifdef OLD_GSOAP
void HTTP_SE::soap_methods(void) {
  if((sp.error = soap_serve_ns__add(&sp)) != SOAP_NO_METHOD) return;
  if((sp.error = soap_serve_ns__update(&sp)) != SOAP_NO_METHOD) return;
  if((sp.error = soap_serve_ns__info(&sp)) != SOAP_NO_METHOD) return;
  if((sp.error = soap_serve_ns__acl(&sp)) != SOAP_NO_METHOD) return;
  if((sp.error = soap_serve_ns__del(&sp)) != SOAP_NO_METHOD) return;
}
#else
void HTTP_SE::soap_methods(void) {
  if(!soap_match_tag(&sp,sp.tag,"ns:add")) soap_serve_ns__add(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"ns:update")) soap_serve_ns__update(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"ns:info")) soap_serve_ns__info(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"ns:acl")) soap_serve_ns__acl(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"ns:del")) soap_serve_ns__del(&sp);
  else sp.error = SOAP_NO_METHOD;
}
#endif

char* soap_strdup_l(soap* sp,const char* s,int l = -1) {
  if(l<0) l=strlen(s);
  char* s_ = (char*)soap_malloc(sp,l+1);
  if(s_ == NULL) return NULL;
  memcpy(s_,s,l); s_[l]=0;
  return s_;
}

void add_path_to_url(string& url,const string& path,bool strip_path = true) {
  int l = url.length()-1;
  if(l == 0) { url=path; return; };
  if(url[l] != '/') url+="/";
  const char* s = path.c_str();
  if(strip_path) if(s[0] == '/') s++;
  url+=s;
  return;
}

/* SOAP calls */
int ns__update(struct soap *sp, ns__fileinfo *file, struct ns__updateResponse &r) {
  // initialize response
  r.error_code=0; r.sub_error_code=0; r.error_description=NULL;
  r.file.id=NULL;
  r.file.size=NULL;
  r.file.checksum=NULL;
  r.file.acl=NULL;
  r.file.url=NULL;
  r.file.__size_url=0;
  r.file.state=NULL;
  r.file.created=NULL;
  //
  // update file attributes and check if state should not be changed 
  //
  if(!file) {
    cerr<<"SOAP: update: missing information about file"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA;
    return SOAP_OK;
  };
  if(file->url) {
    cerr<<"SOAP: update: update does not need contact url"<<endl;
    r.error_code=ERR_EXCESSIVE_INPUT_DATA;
    return SOAP_OK;
  };
  HTTP_SE* it = (HTTP_SE*)(sp->user);
  if(file->id && it->current_name && (*it->current_name) && (strcmp(it->current_name,file->id)!=0)) {
    cerr<<"SOAP: update: if provided file id must be same as path"<<endl;
    r.error_code=ERR_WRONG_INPUT_DATA;
    return SOAP_OK;
  };
  if((!(file->id)) && !((it->current_name) && (*it->current_name))) {
    cerr<<"SOAP: update: missing file id"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA;
    return SOAP_OK;
  };
  if(!(file->id)) file->id=(char*)(it->current_name);
  if(file->id) cerr<<"SOAP: update: file id: "<<file->id<<endl;
  if(file->size) cerr<<"SOAP: update: file size: "<<*(file->size)<<endl;
  if(file->checksum) cerr<<"SOAP: update: checksum: "<<(file->checksum)<<endl;
  if(file->created) cerr<<"SOAP: update: created: "<<(file->created)<<endl;
  if(file->acl) cerr<<"SOAP: update: acl: "<<(file->acl)<<endl;
  if(!(it->current_file)) {
    cerr<<"SOAP: update: there is no file to update"<<endl;
    r.error_code=ERR_INTERNAL_ERROR;
    return SOAP_OK;
  };
  SEFile& f = *(it->current_file);
  // updates accepted from creator only
  if(it->current_file->creator() != it->c->identity().DN()) {
    cerr<<"SOAP: update: non-creator is trying to update"<<endl;
    r.error_code=ERR_INTERNAL_ERROR;
    return SOAP_OK;
  }; 
  bool previous_complete = f.complete();
  if(file->acl) {
    cerr<<"SOAP: update: update does not accept ACL yet"<<endl;
    r.error_code=ERR_EXCESSIVE_INPUT_DATA;
    return SOAP_OK;
  };
  if(file->size) {
    if(f.size() != *(file->size)) {
      cerr<<"SOAP: update: file size differs"<<endl;
      r.error_code=ERR_WRONG_INPUT_DATA;
      return SOAP_OK;
    };
  };
  if(file->checksum) {
    if(f.checksum_available()) {
      if(strcmp(file->checksum,f.checksum().c_str()) != 0) {
        cerr<<"SOAP: update: file checksum differs"<<endl;
        r.error_code=ERR_WRONG_INPUT_DATA;
        return SOAP_OK;
      };
    } else {
      f.checksum(file->checksum);
    };
  };
  if(file->created) {
    if(f.created_available()) {
      if(!f.created_compare(file->created)) {
        cerr<<"SOAP: update: file creation time differs"<<endl;
        r.error_code=ERR_WRONG_INPUT_DATA;
        return SOAP_OK;
      };
    } else {
      f.created(file->created);
      if(!f.created_available()) {
        cerr<<"SOAP: update: file creation uninterpretable"<<endl;
        r.error_code=ERR_WRONG_INPUT_DATA;
        return SOAP_OK;
      };
    };
  };
  if(f.write_attr() != 0) {
    cerr<<"SOAP: update: failed to write attributes"<<endl;
    r.error_code=ERR_INTERNAL_ERROR;
    return SOAP_OK;
  };
// TODO - created
  f.acquire();
  if((!previous_complete) && (f.complete())) { // there is a reason to register
    if(f.state_reg(REG_STATE_REGISTERING)) {
      if(it->files()->NS()->Register(f) != 0) {
        if(it->handle->register_retry()) {
          cerr<<"Failed to register (will retry)"<<endl;
        } else {
          cerr<<"Failed to register (what to do ?)"<<endl;
        };
        f.state_reg(REG_STATE_LOCAL);
      } else {
        f.state_reg(REG_STATE_ANNOUNCED);
      };
    } else {
      if(it->handle->register_retry()) {
        cerr<<"Failed to set REGISTERING state (will retry)"<<endl;
      } else {
        cerr<<"Failed to set REGISTERING state (what to do ?)"<<endl;
      };
    };
  };
  f.release();
  return SOAP_OK;
}

int ns__add(struct soap* sp,ns__fileinfo *file,int __size_source,char** source,
struct ns__addResponse &r) {
  HTTP_SE* it = (HTTP_SE*)(sp->user);

  // initialize response
  r.error_code=0; r.sub_error_code=0; r.error_description=NULL;
  r.file.id=NULL;
  r.file.size=NULL;
  r.file.checksum=NULL;
  r.file.acl=NULL;
  r.file.url=NULL;
  r.file.__size_url=0;
  r.file.state=NULL;
  r.file.created=NULL;

  if(!it->acl_create) {
    SEFiles* files = it->files();
    if(!files) {
      odlog(0)<<"No files"<<endl;
      r.error_code=ERR_INTERNAL_ERROR; return SOAP_OK;
    };
    int acl_flags = files->check_acl(it->c->identity());
    if(!(acl_flags & FILE_ACC_CREATE)) {
      r.error_code=ERR_ACCESS;
      r.error_description="Access denied";
      return SOAP_OK;
    };
  };
  
  // check if all necessary attributes provided
  if(!file) {
    cerr<<"SOAP: add: missing information about file"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA; 
    r.error_description="Missing information about file";
    return SOAP_OK;
  };
  if(file->id) cerr<<"SOAP: add: file id: "<<file->id<<endl;
  if(file->size) cerr<<"SOAP: add: file size: "<<*(file->size)<<endl;
  if(file->checksum) cerr<<"SOAP: add: checksum: "<<(file->checksum)<<endl;
  if(file->acl) cerr<<"SOAP: add: acl: "<<(file->acl)<<endl;
  if(file->created) cerr<<"SOAP: add: created: "<<(file->created)<<endl;

  r.file.id=file->id;
  if(file->url) {
    cerr<<"SOAP: add: add does not need contact url in input parameters"<<endl;
    r.error_code=ERR_EXCESSIVE_INPUT_DATA; 
    return SOAP_OK;
  };
  if(it->current_name && (*it->current_name)) {
    cerr<<"SOAP: add: add requires only service in path"<<endl;
    r.error_code=ERR_EXCESSIVE_INPUT_DATA; 
    return SOAP_OK;
  };
  
  // attributes for response
  r.file.id=file->id;
  r.file.size=file->size;
  r.file.checksum=file->checksum;
  r.file.acl=file->acl;
  if(__size_source) {
    cerr<<"SOAP: add: sources are not supported yet"<<endl;
    r.error_code=ERR_UNSUPPORTED_FEATURE; 
    return SOAP_OK;
  };
  if(!file->id) {
    cerr<<"SOAP: add: missing file id (lfn)"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA; 
    return SOAP_OK;
  };
  if(!file->size) {
    cerr<<"SOAP: add: missing file size (needed for allocation)"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA; 
    return SOAP_OK;
  };
  if(!file->acl) {
    cerr<<"SOAP: add: missing ACL (can't access file without ACL)"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA; 
    return SOAP_OK;
  };
  // create new file
  SEAttributes attr(file->id,*(file->size),it->c->identity());
  if(file->checksum) attr.checksum(string(file->checksum));
  if(file->created) attr.created(file->created);
  if((!attr.complete()) && (it->handle->register_immediately())) {
    cerr<<"SOAP: add: missing file information needed for registration"<<endl;
    r.error_code=ERR_MISSING_INPUT_DATA; 
    return SOAP_OK;
  };
  SEFile* f = it->new_file(attr);
  if(f==NULL) {
    cerr<<"File not created: "<<file->id<<endl;
    r.error_code=ERR_CREATING_NEW_FILE;  // TODO: check for reason of failure
    return SOAP_OK;
  };
  string url_ = it->url;
  add_path_to_url(url_,file->id,false);
  cerr<<"SOAP: add: contact url: "<<url_<<endl;
  r.file.url=(char**)soap_malloc(sp,sizeof(char*)*2); // 2 urls
  r.file.__size_url=0;
  if(r.file.url == NULL) {
    cerr<<"SOAP: add: soap_malloc failed"<<endl;
    f->destroy();
    r.error_code=ERR_INTERNAL_ERROR; 
    return SOAP_OK;
  }; 
  r.file.url[0]=(char*)soap_strdup_l(sp,url_.c_str(),url_.length()+1);
  if(r.file.url[0] == NULL) {
    cerr<<"SOAP: add: soap_strdup_l failed"<<endl;
    f->destroy();
    r.error_code=ERR_INTERNAL_ERROR; 
    return SOAP_OK;
  }; 
  r.file.__size_url=1;
  // hack
  if(strncmp(base_gsi_url.c_str(),r.file.url[0],base_gsi_url.length()) == 0) {
    string u(r.file.url[0]);
    u.replace(0,base_gsi_url.length(),base_ssl_url);
    r.file.url[1]=(char*)soap_strdup_l(sp,u.c_str(),u.length()+1);
    if(r.file.url[1] == NULL) {
      cerr<<"SOAP: add: soap_strdup_l failed"<<endl;
      f->destroy();
      r.error_code=ERR_INTERNAL_ERROR; 
      return SOAP_OK;
    }; 
    r.file.__size_url=2;
  };
  if(f->write_acl(it->c->identity(),file->acl) != 0) {
    cerr<<"SOAP: add: failed to acquire/store ACL"<<endl;
    f->destroy();
    r.error_code=ERR_INTERNAL_ERROR; 
    return SOAP_OK;
  };
  // register new file
  if(it->handle->register_immediately()) { 
    f->acquire();
    if(!f->state_reg(REG_STATE_REGISTERING)) {
      if(it->handle->register_retry()) { 
        cerr<<"SOAP: add: failed to set REG_STATE_REGISTERING (will retry)"<<endl;
        f->state_reg(REG_STATE_LOCAL);
      } else {
        cerr<<"SOAP: add: failed to set REG_STATE_REGISTERING"<<endl;
        f->release();
        f->destroy();
        r.error_code=ERR_INTERNAL_ERROR; 
        return SOAP_OK;
      };
    } else {
      if(it->files()->NS()->Register(*f) != 0) {
        if(it->handle->register_retry()) { 
          cerr<<"SOAP: add: failed to register (will retry)"<<endl;
          f->state_reg(REG_STATE_LOCAL);
        } else {
          cerr<<"SOAP: add: failed to register"<<endl;
          f->release();
          f->destroy();
          r.error_code=ERR_INTERNAL_ERROR; 
          return SOAP_OK;
        };
      } else {
        f->state_reg(REG_STATE_ANNOUNCED);
      };
    };
    f->release();
  } else {
    se_thread->attention();
  };
  int fst = FILE_STATE_COLLECTING;
  if(*(file->size) == 0) fst=FILE_STATE_COMPLETE;
  if(!f->state_file(fst)) {
    cerr<<"SOAP: add: failed to set FILE_STATE_COLLECTING/COMPLETE"<<endl;
    it->files()->NS()->Unregister(*f);  // TODO: if failed set state to repeat later
    f->destroy();
    r.error_code=ERR_INTERNAL_ERROR; 
    return SOAP_OK;
  };
  return SOAP_OK;
}

/* report information about file. */
int ns__info(soap *sp,char *pattern, ns__infoResponse &r) {
  // initialize response
  r.error_code=0; r.sub_error_code=0; r.error_description=NULL;
  r.__size_file=0; r.file=NULL;
  HTTP_SE* it = (HTTP_SE*)(sp->user);
  SEFiles* files = it->files();
  if(!files) {
    odlog(0)<<"No files"<<endl;
    r.error_code=ERR_INTERNAL_ERROR; return SOAP_OK;
  };
  int acl_top_flags = files->check_acl(it->c->identity());
  if(pattern) {
    string pat = pattern;
    if(pat.length() == 0) {  // create pattern from url
      pat="^"; pat+=it->current_name; pat+=".*";
    };
    regex_t preg;
    cerr<<"regex: "<<pat<<endl;
    if(regcomp(&preg,pat.c_str(),REG_EXTENDED | REG_NOSUB) != 0) {
      odlog(0)<<"Failed to compile regex: "<<pat<<endl;
      r.error_code=ERR_INTERNAL_ERROR; return SOAP_OK;
    };
    SEFiles::iterator f;
    files->acquire();
    f=files->begin();
    int n_alloc = 0;
    int n = 0;
    for(;f!=files->end();++f) {
      cerr<<"file: "<<f->id()<<endl;
      if(regexec(&preg,f->id(),0,NULL,0) == 0) {
        cerr<<"matched"<<endl;
        if((it->acl_read) || (acl_top_flags & FILE_ACC_LIST) || (f->check_acl(it->c->identity()) & FILE_ACC_LIST)){
          cerr<<"allowed"<<endl;
          if(n>=n_alloc) {
            n_alloc+=10;
            ns__fileinfo* f_tmp = r.file;
            r.file=(ns__fileinfo*)soap_malloc(sp,sizeof(ns__fileinfo)*n_alloc);
            if(f_tmp != NULL) soap_dealloc(sp,f_tmp);
            if(r.file == NULL) {
              r.error_code=ERR_INTERNAL_ERROR; n=0; n_alloc=0; break;
            };
            if(f_tmp != NULL) memcpy(r.file,f_tmp,sizeof(ns__fileinfo)*n);
          };
          ns__fileinfo rr;
          rr.id=soap_strdup_l(sp,f->id());
          rr.size=NULL;
          rr.checksum=NULL;
          rr.acl=NULL;
          rr.__size_url=0;
          rr.url=NULL;
          rr.created=NULL;
          rr.state=NULL;
          memcpy(r.file+n,&rr,sizeof(ns__fileinfo));
          n++;
        };
      };
    };
    files->release();
    r.__size_file=n;
    regfree(&preg);
    return SOAP_OK;
  } else {
    if(!(it->current_file)) {
      cerr<<"SOAP: info: there is no such file"<<endl;
      r.error_code=ERR_INTERNAL_ERROR;
      return SOAP_OK;
    };
    if((it->acl_read) || (acl_top_flags & FILE_ACC_LIST) || (it->current_file->check_acl(it->c->identity()) & FILE_ACC_LIST)) {
      cerr<<"allowed"<<endl;
      r.file=(ns__fileinfo*)soap_malloc(sp,sizeof(ns__fileinfo));
      if(r.file == NULL) { r.error_code=ERR_INTERNAL_ERROR; return SOAP_OK; };
      ns__fileinfo rr;
      rr.id=soap_strdup_l(sp,it->current_name);
      rr.size=(uint64_t*)soap_malloc(sp,sizeof(uint64_t));
      if(rr.size) (*rr.size)=it->current_file->size();
      rr.checksum=soap_strdup_l(sp,it->current_file->checksum().c_str());
      string acl;
      if(it->current_file->read_acl(it->c->identity(),acl) == 0) {
        rr.acl=soap_strdup_l(sp,acl.c_str(),acl.length());
      };
      string url_ = it->url + "/" + it->current_file->id();
      rr.__size_url=0;
      rr.url=(char**)soap_malloc(sp,sizeof(char*)*2); // 2 urls
      if(rr.url) {
        rr.url[0]=(char*)soap_strdup_l(sp,url_.c_str(),url_.length()+1);
        if(rr.url[0]) {
          rr.__size_url=1;
          if(strncmp(base_gsi_url.c_str(),rr.url[0],base_gsi_url.length()) == 0) {
            string u(rr.url[0]);
            u.replace(0,base_gsi_url.length(),base_ssl_url);
            rr.url[1]=(char*)soap_strdup_l(sp,u.c_str(),u.length()+1);
            if(rr.url[1]) rr.__size_url=2;
          };
        };
      };
      rr.state=(ns__filestate*)soap_malloc(sp,sizeof(ns__filestate));
      if(rr.state) (*rr.state)=(ns__filestate)(it->current_file->state_file());
      memcpy(r.file,&rr,sizeof(ns__fileinfo));
      r.__size_file=1;
    } else {
      r.error_code=ERR_ACCESS;
    };
  };
  return SOAP_OK;
}

/* report/modify ACL */
int ns__acl(soap *sp, char *acl, ns__aclResponse &r) {
  int res = -1;
  r.error_code=SUCCESS; r.error_description=NULL;
  r.sub_error_code=0;
  r.acl=NULL;
  HTTP_SE* it = (HTTP_SE*)(sp->user);
  if(it->current_file == NULL) {
    // Top ACL
    SEFiles* files = it->files();
    if(!files) {
      odlog(0)<<"No files"<<endl;
      r.error_code=ERR_INTERNAL_ERROR; return SOAP_OK;
    };
    int acl_top_flags = files->check_acl(it->c->identity());
    if(acl) {
      if(!(acl_top_flags & FILE_ACC_WRITE_META)) {
        cerr<<"SOAP: acl: inot allowed to write acl"<<endl;
        r.error_code=ERR_INSUFFICIENT_ACCESS;
        return SOAP_OK;
      };
      res=files->write_acl(it->c->identity(),acl);
    } else {
      if(!(acl_top_flags & FILE_ACC_READ_META)) {
        cerr<<"SOAP: acl: not allowed to write acl"<<endl;
        r.error_code=ERR_INSUFFICIENT_ACCESS;
        return SOAP_OK;
      };
      string acl_;
      res=it->current_file->read_acl(it->c->identity(),acl_);
      if(res==0) {
        if((r.acl=(char*)soap_malloc(sp,acl_.length()+1)) == NULL) {
          r.error_code=ERR_INTERNAL_ERROR;
          return SOAP_OK;
        };
        strcpy(r.acl,acl_.c_str());
      };
    };    
  } else {
    int acl_flags = it->current_file->check_acl(it->c->identity());
    if(acl) {
      if(acl_flags & FILE_ACC_WRITE_META) {
        res=it->current_file->write_acl(it->c->identity(),acl);
      } else {
        r.error_code=ERR_INSUFFICIENT_ACCESS;
      };
    } else {
      if(acl_flags & FILE_ACC_READ_META) {
        string acl_;
        res=it->current_file->read_acl(it->c->identity(),acl_);
        if(res==0) {
          if((r.acl=(char*)soap_malloc(sp,acl_.length()+1)) == NULL) {
            r.error_code=ERR_INTERNAL_ERROR;
            return SOAP_OK;
          };
          strcpy(r.acl,acl_.c_str());
        };
      } else {
        r.error_code=ERR_INSUFFICIENT_ACCESS;
      };
    };
  };
  if((res!=0) && (r.error_code==SUCCESS)) {
    r.error_code=ERR_INTERNAL_ERROR;
  };
  return SOAP_OK;
}

int ns__del(soap *sp, ns__delResponse &r) {
  // initialize response
  r.error_code=0; r.sub_error_code=0; r.error_description=NULL;
  HTTP_SE* it = (HTTP_SE*)(sp->user);
  if(it->current_file == NULL) {
    cerr<<"SOAP: del: file is missing"<<endl;
    r.error_code=ERR_OBJECT_MISSING;
    return SOAP_OK;
  };
  int acl_flags = it->current_file->check_acl(it->c->identity());
  SEFiles* files = it->files();
  if(!files) {
    odlog(0)<<"No files"<<endl;
    r.error_code=ERR_INTERNAL_ERROR; return SOAP_OK;
  };
  int acl_top_flags = files->check_acl(it->c->identity());
  if((!(acl_flags & FILE_ACC_DELETE)) && (!(acl_top_flags & FILE_ACC_DELETE)) && !it->acl_create) {
    cerr<<"SOAP: del: insufficient access: "<<acl_flags<<" - "<<FILE_ACC_DELETE<<endl;
    r.error_code=ERR_INSUFFICIENT_ACCESS;
    return SOAP_OK;
  };
  SEFile* f = it->current_file;
  // analyze current state
  f->acquire();
  if(f->state_file() == FILE_STATE_DELETING) { // already
    cerr<<"SOAP: del: already deleting"<<endl;
    f->release(); return SOAP_OK;
  };
  if(!f->state_file(FILE_STATE_DELETING)) {
    cerr<<"SOAP: del: failed to change file state"<<endl;
    r.error_code=ERR_INTERNAL_ERROR;
    f->release(); return SOAP_OK;
  };
/* 
Race condition here. After release() unregistering thread can 
try to remove file. In a future all SEFile* will be replaced
with SEFiles::iterator  
  if(f->state_reg() == REG_STATE_LOCAL) { // ready to by wiped out
    cerr<<"SOAP: del: removing directly"<<endl;
    f->release(); it->files()->remove(*f) return SOAP_OK;
  };
*/
  bool to_unregister = false;
  if((f->state_reg() == REG_STATE_ANNOUNCED) &&
     (it->handle->register_immediately())) {
    if(!f->state_reg(REG_STATE_UNREGISTERING)) {
      if(it->handle->register_retry()) {
        cerr<<"SOAP: del: failed to set REG_STATE_UNREGISTERING (will retry)"<<endl;
      } else {
        cerr<<"SOAP: del: failed to set REG_STATE_UNREGISTERING"<<endl; 
        // ???????????
        r.error_code=ERR_INTERNAL_ERROR;
        f->release(); return SOAP_OK;
      };
    } else {
      to_unregister=true;
    };
  };
  f->release();
  if(to_unregister) {
    if(it->files()->NS()->Unregister(*f) != 0) {
      f->state_reg(REG_STATE_ANNOUNCED);
      if(it->handle->register_retry()) {
        cerr<<"SOAP: del: failed to unregister (will retry)"<<endl;
      } else {
        cerr<<"SOAP: del: failed to unregister"<<endl;
        r.error_code=ERR_INTERNAL_ERROR;
        return SOAP_OK;
      };
      to_unregister=false;
    } else {
      f->state_reg(REG_STATE_LOCAL);
      it->files()->remove(*f);
    };
  };
  if(!to_unregister) {
    se_thread->attention();
  };
  return SOAP_OK;
}

